/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.objectweb.petals.component.framework.listener;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.logging.Logger;

import org.objectweb.petals.component.framework.AbstractComponent;
import org.objectweb.petals.component.framework.interceptor.Interceptor;
import org.objectweb.petals.component.framework.interceptor.MessageInterceptorManager;
import org.objectweb.petals.component.framework.interceptor.description.InterceptorConfiguration;
import org.objectweb.petals.component.framework.su.ServiceUnitDataHandler;
import org.objectweb.petals.component.framework.su.ServiceUnitManager;
import org.objectweb.petals.component.framework.util.Exchange;
import org.objectweb.petals.component.framework.util.ExchangeImpl;
import org.objectweb.petals.jbi.descriptor.ServiceUnitExtensibleElement;

/**
 * The message interceptor processor. Used by the component to intercept
 * messages.
 *
 * @author Christophe HAMERLING (chamerling) - eBM WebSourcing
 *
 */
public class MessageInterceptorProcessor {

    enum Phase {
        SEND, ACCEPT;
    }

    private static final Map<String, String> EMPTY_PARAMETERS = new HashMap<String, String>();

    /**
     * The JBI component instance
     */
    protected AbstractComponent component;

    /**
     * The logger
     */
    protected Logger logger;

    /**
     * Creates a new instance of {@link MessageInterceptorProcessor}
     *
     */
    public MessageInterceptorProcessor(AbstractComponent component,
            Logger logger) {
        this.component = component;
        this.logger = logger;
    }

    /**
     * Intercepts the message.
     *
     * @param exchange
     * @param extensible
     */
    public void intercept(final Exchange exchange,
            final ServiceUnitExtensibleElement extensible, final Phase phase) {
        logger.fine("Intercepting message processing on " + phase);

        ServiceUnitManager suManager = (ServiceUnitManager) component
                .getServiceUnitManager();
        ServiceUnitDataHandler dataHandler = suManager
                .getSuDataHandlerForConsumeProvide(extensible);

        if (dataHandler != null) { // no DH when no deployed SU
            MessageInterceptorManager messageInterceptorManager = dataHandler
                    .getMessageInterceptorManager(extensible);

            if (messageInterceptorManager != null) {
                // unactivate the check on the Role when intercepting at
                // reception,
                // to be able to reset the normalized message.
                if (Phase.ACCEPT.equals(phase)) {
                    ((ExchangeImpl) exchange).setCheckRole(false);
                }
                this.intercept(exchange, extensible, messageInterceptorManager,
                        phase);
                if (Phase.ACCEPT.equals(phase)) {
                    ((ExchangeImpl) exchange).setCheckRole(true);
                }
            }
        }
    }

    /**
     *
     * @param exchange
     * @param extensible
     * @param messageInterceptorManager
     */
    protected void intercept(Exchange exchange,
            ServiceUnitExtensibleElement extensible,
            MessageInterceptorManager messageInterceptorManager, Phase phase) {
        Map<String, Interceptor> interceptors = null;
        Map<String, InterceptorConfiguration> additionalConfiguration = null;

        if (messageInterceptorManager != null) {

            // getting good interceptors
            if (phase == Phase.ACCEPT) {
                interceptors = messageInterceptorManager
                        .getInputOrderedInterceptors();
                additionalConfiguration = messageInterceptorManager
                        .getInputAdditionalConfiguration();
            } else if (phase == Phase.SEND) {
                interceptors = messageInterceptorManager
                        .getOutputOrderedInterceptors();
                additionalConfiguration = messageInterceptorManager
                        .getOutputAdditionalConfiguration();
            }

            Iterator<String> iterator = interceptors.keySet().iterator();
            while (iterator.hasNext()) {
                String interceptorName = iterator.next();
                logger.fine("Calling message interceptor " + interceptorName);

                // get interceptor
                Interceptor interceptor = interceptors.get(interceptorName);

                // get additional configuration from consume/provide
                InterceptorConfiguration configuration = additionalConfiguration
                        .get(interceptorName);
                Map<String, String> parameters = null;
                if (configuration != null) {
                    parameters = configuration.getParameters();
                } else {
                    // empty
                    parameters = EMPTY_PARAMETERS;
                }

                // override component parameters with SU ones
                Map<String, String> finalParameters = new HashMap<String, String>();
                finalParameters.putAll(interceptor.getComponentParameters());
                finalParameters.putAll(parameters);

                // intercept...
                if (phase == Phase.ACCEPT) {
                    logger.fine("Intercepting message on accept phase");
                    interceptor.handleMessageAccept(exchange, extensible,
                            finalParameters);
                } else if (phase == Phase.SEND) {
                    logger.fine("Intercepting message on send phase");
                    interceptor.handleMessageSent(exchange, extensible,
                            finalParameters);
                }
            }

        } else {
            logger.warning("No interceptor manager has been found");
        }
    }
}
