/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2005 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.objectweb.petals.component.framework.su;

import java.io.File;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.JBIException;
import javax.jbi.management.DeploymentException;
import javax.jbi.servicedesc.ServiceEndpoint;
import javax.xml.namespace.QName;

import org.objectweb.petals.component.framework.AbstractComponent;
import org.objectweb.petals.component.framework.ExtensionsConstants;
import org.objectweb.petals.component.framework.PEtALSCFException;
import org.objectweb.petals.component.framework.interceptor.MessageInterceptorManager;
import org.objectweb.petals.component.framework.interceptor.description.ExtensibleElementInterceptorConfiguration;
import org.objectweb.petals.component.framework.interceptor.description.InterceptorBuilder;
import org.objectweb.petals.component.framework.util.Extensions;
import org.objectweb.petals.component.framework.util.IDGenerator;
import org.objectweb.petals.component.framework.util.ManagementMessageUtil;
import org.objectweb.petals.component.framework.util.PropertiesUtil;
import org.objectweb.petals.component.framework.util.WSDLHelper;
import org.objectweb.petals.jbi.descriptor.Consumes;
import org.objectweb.petals.jbi.descriptor.JBIDescriptor;
import org.objectweb.petals.jbi.descriptor.JBIDescriptorBuilder;
import org.objectweb.petals.jbi.descriptor.JBIDescriptorException;
import org.objectweb.petals.jbi.descriptor.Provides;
import org.objectweb.petals.jbi.descriptor.ServiceUnitExtensibleElement;
import org.w3c.dom.Document;

/**
 * A simple ServiceUnitManager. It processes service unit jbi descriptor : <br/>
 * Provides nodes : new endpoints registered in JBI container<br/> Consumes
 * nodes : link addresses to existing JBI endpoints
 *
 * @author ofabre,alouis
 *
 */
public class ServiceUnitManager implements javax.jbi.component.ServiceUnitManager {

    private static final String JBI_XML = "jbi.xml";

    private static final String META_INF = "META-INF";

    /**
     * The component's Logger
     */
    private Logger logger; // NOPMD by gblondelle

    /**
     * The service unit data handlers. Each service unit is linked to a
     * datahander. Key is the service unit name.
     */
    private final Map<String, ServiceUnitDataHandler> serviceUnitDataHandlers;

    protected final AbstractComponent component;

    /**
     * Creates a new instance of {@link ServiceUnitManager}
     *
     * @param component
     */
    public ServiceUnitManager(AbstractComponent component) {
        super();
        this.component = component;
        this.logger = component.getLogger();
        this.serviceUnitDataHandlers = new HashMap<String, ServiceUnitDataHandler>();
    }

    /**
     * (non-Javadoc)
     *
     * @see javax.jbi.component.ServiceUnitManager#deploy(java.lang.String,
     *      java.lang.String)
     */
    public String deploy(final String serviceUnitName, final String suRootPath)
            throws DeploymentException {
        logger.log(Level.FINE, "deploy  serviceUnitName  " + serviceUnitName
                + "serviceUnitRootPath" + suRootPath);

        if (serviceUnitDataHandlers.get(serviceUnitName) == null) {
            ServiceUnitDataHandler handler = new ServiceUnitDataHandler();
            handler.setInstallRoot(suRootPath);
            handler.setName(serviceUnitName);
            File jbiXmlFile = new File(suRootPath + File.separator + META_INF + File.separator
                    + JBI_XML);

            JBIDescriptorBuilder builder = null;
            JBIDescriptor descriptor = null;
            try {
                builder = JBIDescriptorBuilder.getInstance();
            } catch (JBIDescriptorException e) {
                String msg = "Can't load jbi.xml schema";
                throw new DeploymentException(msg, e);
            }
            try {
                descriptor = builder.validateAndBuildDescriptor(jbiXmlFile.toURI());
            } catch (JBIDescriptorException e) {
                String msg = "Bad JBI Descriptor";
                throw new DeploymentException(msg, e);
            }
            handler.setDescriptor(descriptor);

            // validate the SU extensions if validator is available
            if (component.getServiceUnitExtensionsValidator() != null) {
                try {
                    validateExtensions(descriptor);
                } catch (PEtALSCFException e) {
                    String msg = "Failed to validate at least one SU extensions";
                    throw new DeploymentException(msg, e);
                }
            }

            try {
                processConsumesNodes(handler);
            } catch (JBIException ex) {
                String msg = "Failed to process consumes nodes";
                throw new DeploymentException(msg, ex);
            }

            logger.log(Level.FINEST, "Register a data handler for su : " + serviceUnitName);
            serviceUnitDataHandlers.put(serviceUnitName, handler);

            // additional deployment by component implementation
            if (component.getServiceUnitListener() != null) {
                try {
                    component.getServiceUnitListener().onSUDeployed(serviceUnitName, suRootPath,
                            descriptor);
                } catch (PEtALSCFException e) {
                    // clean cache
                    logger.log(Level.FINEST, "Unregister a data handler for su : "
                            + serviceUnitName);
                    serviceUnitDataHandlers.remove(serviceUnitName);

                    throw new DeploymentException("The ServiceUnit listener failed to process.", e);
                }
            }
        } else {
            throw new DeploymentException(
                    "A service unit with the given name is already registered");
        }
        String result = ManagementMessageUtil.getComponentTaskResult(component.getContext()
                .getComponentName(), "deploy", ManagementMessageUtil.TASK_RESULT_SUCCESS);
        return result;
    }

    /**
     * Retrieve the Provides node that has activated the given jbi Endpoint
     *
     * @param serviceEndpoint
     *            the previously activated jbi Endpoint
     * @return the {@link Provides} element
     */
    public Provides getProvidesFromEndpoint(final ServiceEndpoint serviceEndpoint) {
        Provides result = null;
        for (Iterator<ServiceUnitDataHandler> iter = serviceUnitDataHandlers.values().iterator(); iter
                .hasNext()
                && result == null;) {
            ServiceUnitDataHandler handler = iter.next();
            Map<ServiceEndpoint, Provides> epJBIDescs = handler.getEpJBIDesc();
            result = epJBIDescs.get(serviceEndpoint);
        }
        return result;
    }

    /**
     * Retrieve the Consumes node matching the given destination information
     * (endpoint name, service name, interface name). Search order is by
     * ascending priority : endpointName, serviceName, interfaceName. If there
     * are many consumes matching information, return the first one found.
     *
     * @param endpointName
     *            an endpoint name, can be null
     * @param serviceName
     *            a service {@link QName}, can be null
     * @param interfaceName
     *            an interface {@link QName}, can be null
     * @return the {@link Consumes} element matching information, can be null
     */
    public Consumes getConsumesFromDestination(String endpointName, QName serviceName,
            QName interfaceName) {
        Consumes result = null;
        if (endpointName != null) {
            result = searchConsumesMatchingEndpoint(endpointName);
        }
        if (result == null) {
            if (serviceName != null) {
                result = searchConsumesMatchingService(serviceName);
            }
            if (result == null) {
                if (interfaceName != null) {
                    result = searchConsumesMatchingInterface(interfaceName);
                }
            }
        }
        return result;
    }

    /**
     *
     * @param endpointName
     * @return
     */
    private Consumes searchConsumesMatchingEndpoint(String endpointName) {
        Consumes result = null;
        for (Iterator<ServiceUnitDataHandler> iter = serviceUnitDataHandlers.values().iterator(); iter
                .hasNext()
                && result == null;) {
            ServiceUnitDataHandler handler = iter.next();
            List<Consumes> consumesList = handler.getConsumesList();

            for (Iterator<Consumes> iterator = consumesList.iterator(); iterator.hasNext()
                    && result == null;) {
                Consumes consumes = iterator.next();
                String epName = consumes.getEndpointName();
                if (endpointName.equals(epName)) {
                    result = consumes;
                }
            }
        }
        return result;
    }

    /**
     *
     * @param serviceName
     * @return
     */
    private Consumes searchConsumesMatchingService(QName serviceName) {
        Consumes result = null;
        for (Iterator<ServiceUnitDataHandler> iter = serviceUnitDataHandlers.values().iterator(); iter
                .hasNext()
                && result == null;) {
            ServiceUnitDataHandler handler = iter.next();

            List<Consumes> consumesList = handler.getConsumesList();
            for (Iterator<Consumes> iterator = consumesList.iterator(); iterator.hasNext()
                    && result == null;) {
                Consumes consumes = iterator.next();
                QName servName = consumes.getServiceName();
                if (serviceName.equals(servName)) {
                    result = consumes;
                }
            }
        }
        return result;
    }

    /**
     *
     * @param interfaceName
     * @return
     */
    private Consumes searchConsumesMatchingInterface(QName interfaceName) {
        Consumes result = null;
        for (Iterator<ServiceUnitDataHandler> iter = serviceUnitDataHandlers.values().iterator(); iter
                .hasNext()
                && result == null;) {
            ServiceUnitDataHandler handler = iter.next();

            List<Consumes> consumesList = handler.getConsumesList();
            for (Iterator<Consumes> iterator = consumesList.iterator(); iterator.hasNext()
                    && result == null;) {
                Consumes consumes = iterator.next();
                QName intName = consumes.getInterfaceName();
                if (interfaceName.equals(intName)) {
                    result = consumes;
                }
            }
        }
        return result;
    }

    /**
     * Retrieve the Service description (as Dom Document) of the given jbi
     * Endpoint
     *
     * @param serviceEndpoint
     *            the jbi Endpoint
     * @return the service description as Dom Document
     */
    public Document getServiceDescription(final ServiceEndpoint serviceEndpoint) {
        Document result = null;
        for (Iterator<ServiceUnitDataHandler> iter = serviceUnitDataHandlers.values().iterator(); iter
                .hasNext()
                && result == null;) {
            ServiceUnitDataHandler handler = iter.next();
            Map<ServiceEndpoint, Document> epServDescs = handler.getEpServiceDesc();
            result = epServDescs.get(serviceEndpoint);
        }
        return result;
    }

    /*
     * (non-Javadoc)
     *
     * @see javax.jbi.component.ServiceUnitManager#init(java.lang.String,
     *      java.lang.String)
     */
    public void init(final String serviceUnitName, final String suRootPath)
            throws DeploymentException {
        logger.log(Level.FINE, "init  serviceUnitName  " + serviceUnitName
                + "  serviceUnitRootPath " + suRootPath);
    }

    /*
     * (non-Javadoc)
     *
     * @see javax.jbi.component.ServiceUnitManager#shutDown(java.lang.String)
     */
    public void shutDown(final String serviceUnitName) throws DeploymentException {
        logger.log(Level.FINE, "shutdown  serviceUnitName  " + serviceUnitName);
    }

    /*
     * (non-Javadoc)
     *
     * @see javax.jbi.component.ServiceUnitManager#start(java.lang.String)
     */
    public void start(final String serviceUnitName) throws DeploymentException {
        logger.log(Level.FINE, "start serviceUnitName  " + serviceUnitName);
        ServiceUnitDataHandler handler = serviceUnitDataHandlers.get(serviceUnitName);
        try {
            processProvidesNodes(handler);
        } catch (JBIException ex) {
            String msg = "Failed to start service unit : " + serviceUnitName;
            logger.log(Level.SEVERE, msg, ex);
            throw new DeploymentException(msg, ex);
        }
        if (component.getServiceUnitListener() != null) {
            try {
                component.getServiceUnitListener().onSUStarted(serviceUnitName);
            } catch (PEtALSCFException e) {
                throw new DeploymentException("The ServiceUnit listener failed to process.", e);
            }
        }
    }

    /*
     * (non-Javadoc)
     *
     * @see javax.jbi.component.ServiceUnitManager#stop(java.lang.String)
     */
    public void stop(final String serviceUnitName) throws DeploymentException {
        if (component.getServiceUnitListener() != null) {
            try {
                component.getServiceUnitListener().onSUStopped(serviceUnitName);
            } catch (PEtALSCFException e) {
                throw new DeploymentException("The ServiceUnit listener failed to process.", e);
            }
        }
        try {
            ServiceUnitDataHandler handler = serviceUnitDataHandlers.get(serviceUnitName);
            List<ServiceEndpoint> endpoints = getServiceUnitEndpoints(serviceUnitName);
            for (ServiceEndpoint endpoint : endpoints) {
                component.getContext().deactivateEndpoint(endpoint);
                handler.removeEndpoint(endpoint);
            }

        } catch (JBIException e) {
            throw new DeploymentException("Deactivate endpoint failed", e);
        }
        logger.log(Level.FINE, "stop  serviceUnitName  " + serviceUnitName);
    }

    /*
     * (non-Javadoc)
     *
     * @see javax.jbi.component.ServiceUnitManager#undeploy(java.lang.String,
     *      java.lang.String)
     */
    public String undeploy(final String serviceUnitName, final String serviceUnitRootPath)
            throws DeploymentException {
        ServiceUnitDataHandler handler = serviceUnitDataHandlers.get(serviceUnitName);

        if (component.getServiceUnitListener() != null) {
            try {
                component.getServiceUnitListener().onSUUndeployed(serviceUnitName,
                        serviceUnitRootPath, handler.getDescriptor());
            } catch (PEtALSCFException e) {
                throw new DeploymentException("The ServiceUnit listener failed to process.", e);
            }
        }
        logger.log(Level.FINEST, "Unregister a data handler for su : " + serviceUnitName);
        serviceUnitDataHandlers.remove(serviceUnitName);
        return ManagementMessageUtil.getComponentTaskResult(component.getContext()
                .getComponentName(), "undeploy", ManagementMessageUtil.TASK_RESULT_SUCCESS);
    }

    /**
     * Search the wsdl description of the activated endpoint in the Provides
     * extensions. If no wsdl description can be found, it creates a light
     * weight Wsdl description from Provides Interface, Service and Endpoint
     * names.
     *
     * @param installRoot
     *            the SU install root
     * @param provides
     *            the Provices node
     * @return the WSDL description as Dom Document
     */
    @SuppressWarnings("unchecked")
    protected Document getEnpointWSDLDesc(String installRoot, Provides provides)
            throws PEtALSCFException {
        Document wsdlDesc = null;

        Extensions extensions = new Extensions(provides.getExtensions());

        // Now try to replaced optionnal place holders by real external
        // property values
        String propertiesFilePath = component.getComponentConfiguration().getProperties()
                .getProperty(ExtensionsConstants.PROPERTIES_FILE);
        try {
            PropertiesUtil.fillPlaceHolder(propertiesFilePath, extensions.getPetalsExtensions());
        } catch (PEtALSCFException e2) {
            getLogger().log(Level.WARNING, "Cannot fill place holder of SU extensions", e2);
        }

        if (extensions != null) {
            String wsdlFileLocation = extensions.getPetalsExtensions().getProperty(
                    ExtensionsConstants.WSDL_LOCATION);

            if (wsdlFileLocation != null) {
                boolean urlLocation = true;
                URL url = null;
                try {
                    url = new URL(wsdlFileLocation);
                } catch (MalformedURLException e1) {
                    urlLocation = false;
                }

                if (urlLocation) {
                    try {
                        wsdlDesc = WSDLHelper.createDocumentFromWSDL(url);
                    } catch (Exception e) {
                        throw new PEtALSCFException("Can not retrieve WSDL document from URL : "
                                + url);
                    }
                } else {
                    // try to get the relative path
                    if (!wsdlFileLocation.startsWith(File.separator)) {
                        wsdlFileLocation = File.separator + wsdlFileLocation;
                    }

                    File wsdlFile = new File(installRoot + wsdlFileLocation);
                    if (!wsdlFile.exists()) {
                        throw new PEtALSCFException(
                                "Can not retrieve WSDL document from location : "
                                        + wsdlFileLocation);
                    }
                    wsdlDesc = WSDLHelper.createDocumentFromWSDL(wsdlFile);
                }
            }
        }

        // if the wsdl description is null, generate it
        if (wsdlDesc == null) {
            wsdlDesc = WSDLHelper.createLightWSDL20(provides.getInterfaceName(), provides
                    .getServiceName(), provides.getEndpointName());
        }

        return wsdlDesc;
    }

    /**
     *
     * @param handler
     * @param consumes
     * @throws JBIException
     */
    protected void processConsumesNode(final ServiceUnitDataHandler handler, final Consumes consumes)
            throws JBIException {

        if (consumes != null) {
            handler.addConsumes(consumes);
        }

        this.processInterceptors(handler, consumes);
    }

    /**
     * Process the interceptors for a consume/provide node.
     *
     * @param handler
     * @param consumes
     * @throws
     */
    protected void processInterceptors(final ServiceUnitDataHandler handler,
            final ServiceUnitExtensibleElement suExtensibleElement) throws JBIException {
        Extensions extensions = new Extensions(suExtensibleElement.getExtensions());
        ExtensibleElementInterceptorConfiguration configuration = loadInterceptorConfigurationFromSU(extensions);

        // create the manager for this extension element. Each element has a
        // linked manager with preloaded interceptors.
        if (configuration != null) {
            MessageInterceptorManager interceptorManager = new MessageInterceptorManager(
                    getLogger());
            interceptorManager.init(component.getInterceptors(), configuration
                    .getInputInterceptors(), configuration.getOutputInterceptors());

            handler.addMessageInterceptorManager(suExtensibleElement, interceptorManager);
        }
    }

    /**
     *
     * @param handler
     * @throws JBIException
     */
    protected void processConsumesNodes(final ServiceUnitDataHandler handler) throws JBIException {
        JBIDescriptor descriptor = handler.getDescriptor();
        List<Consumes> consumesList = descriptor.getServices().getConsumes();
        for (Consumes consumes : consumesList) {
            processConsumesNode(handler, consumes);
        }
    }

    /**
     *
     * @param extensions
     * @return an interceptor configuration or null if no interceptor has been
     *         defined in the extensions.
     */
    protected ExtensibleElementInterceptorConfiguration loadInterceptorConfigurationFromSU(
            Extensions extensions) {
        InterceptorBuilder builder = InterceptorBuilder.getInstance();
        return builder.loadSUInterceptor(extensions);
    }

    /**
     *
     * @param handler
     * @param provides
     * @return
     * @throws JBIException
     */
    protected ServiceEndpoint processProvidesNode(final ServiceUnitDataHandler handler,
            final Provides provides) throws JBIException {

        String endpointName = provides.getEndpointName();

        if (ExtensionsConstants.AUTOGENERATED_ENDPOINT_NAME.equals(endpointName)) {
            endpointName = IDGenerator.getNewID();
        }

        ServiceEndpoint ep = component.getContext().activateEndpoint(provides.getServiceName(),
                endpointName);
        try {
            handler.addEndpointDescription(ep, getEnpointWSDLDesc(handler.getInstallRoot(),
                    provides));
        } catch (JBIException e) {
            // deactive the endpoint as the start process has failed
            component.getContext().deactivateEndpoint(ep);
            throw e;
        }
        handler.addJBIDescription(ep, provides);
        this.processInterceptors(handler, provides);

        return ep;
    }

    /**
     *
     * @param handler
     * @throws JBIException
     */
    protected void processProvidesNodes(final ServiceUnitDataHandler handler) throws JBIException {
        JBIDescriptor descriptor = handler.getDescriptor();

        List<Provides> providesList = descriptor.getServices().getProvides();
        for (Provides provides : providesList) {
            processProvidesNode(handler, provides);
        }
    }

    /**
     *
     * @param serviceUnitName
     * @return
     */
    private List<ServiceEndpoint> getServiceUnitEndpoints(final String serviceUnitName) {
        return new ArrayList<ServiceEndpoint>(serviceUnitDataHandlers.get(serviceUnitName)
                .getEpServiceDesc().keySet());
    }

    /**
     * Return all {@link ServiceUnitDataHandler}s as a Map
     *
     * @return all {@link ServiceUnitDataHandler}s as a Map
     */
    public Map<String, ServiceUnitDataHandler> getServiceUnitDataHandlers() {
        return this.serviceUnitDataHandlers;
    }

    /**
     * Return the component Logger
     *
     * @return the component Logger
     */
    public Logger getLogger() {
        return logger;
    }

    /**
     * Retrieve the install root of the Service Unit that activated the given
     * endpoint
     *
     * @param ep
     *            the previously activated endpoint
     * @return the Service unit install root
     */
    public String getSUInstallRootForActivatedEp(ServiceEndpoint ep) {
        ServiceUnitDataHandler handler = getSuDataHandlerForEp(ep);
        return handler.getInstallRoot();
    }

    /**
     * Retrieve the data handler of the Service Unit that activated the given
     * endpoint.
     *
     * @param ep
     *            the previously activated endpoint, or <code>null</code> if
     *            not found.
     * @return the {@link ServiceUnitDataHandler}
     */
    public ServiceUnitDataHandler getSuDataHandlerForEp(final ServiceEndpoint ep) {
        ServiceUnitDataHandler result = null;
        if (ep != null) {
            for (Iterator<ServiceUnitDataHandler> iter = this.serviceUnitDataHandlers.values()
                    .iterator(); iter.hasNext() && result == null;) {
                ServiceUnitDataHandler element = iter.next();

                for (Iterator<ServiceEndpoint> iterator2 = element.getEpJBIDesc().keySet()
                        .iterator(); iterator2.hasNext() && result == null;) {
                    ServiceEndpoint element2 = iterator2.next();
                    if (ep.equals(element2)) {
                        result = element;
                    }
                }
            }
        }
        return result;
    }

    /**
     *
     * @param extensible
     * @return
     */
    public ServiceUnitDataHandler getSuDataHandlerForConsumeProvide(
            final ServiceUnitExtensibleElement extensible) {
        ServiceUnitDataHandler result = null;

        if (extensible != null) {
            for (Iterator<ServiceUnitDataHandler> iter = this.serviceUnitDataHandlers.values()
                    .iterator(); iter.hasNext() && result == null;) {
                ServiceUnitDataHandler element = iter.next();

                for (Iterator<ServiceUnitExtensibleElement> iterator2 = element
                        .getMessageInterceptorManager().keySet().iterator(); iterator2.hasNext()
                        && result == null;) {

                    ServiceUnitExtensibleElement element2 = iterator2.next();
                    if (extensible.equals(element2)) {
                        result = element;
                    }
                }
            }
        }

        return result;
    }

    /**
     * Validate the SU extensions.
     *
     * @param descriptor
     *            the service unit descriptor
     * @throws PEtALSCFException
     *             if at least an extension block is not valid. It is the role
     *             of the developer to specify which extension is not valid.
     */
    protected void validateExtensions(JBIDescriptor descriptor) throws PEtALSCFException {

        // validate the provides
        List<Provides> providesList = descriptor.getServices().getProvides();
        for (Provides provides : providesList) {
            component.getServiceUnitExtensionsValidator().validateProvidesExtensions(
                    provides.getExtensions());
        }

        // validate the consumes
        List<Consumes> consumesList = descriptor.getServices().getConsumes();
        for (Consumes consumes : consumesList) {
            component.getServiceUnitExtensionsValidator().validateConsumesExtensions(
                    consumes.getExtensions());
        }
    }
}
