/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.objectweb.petals.component.framework.util;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLConnection;
import java.util.HashSet;
import java.util.Properties;
import java.util.Set;
import java.util.Map.Entry;

import org.objectweb.petals.component.framework.PEtALSCFException;

/**
 * An util class for properties file management
 *
 * @author ofabre - eBM Websourcing
 *
 */
public class PropertiesUtil {

    /**
     * <p>
     * Load a {@link Properties} object from a properties file path. The file
     * path can be relative path based on Petals root path or a valid URL.
     * </p>
     * <p>
     * Exemples : <br/> Relative path : /conf/external.properties <br/> Absolute
     * path : file:///C:/conf/external.properties <br/> URL :
     * ftp://user:pwd@host:port/external.properties
     * </p>
     *
     * @param propertiesFilePath
     *            The file path can be relative path based on Petals root path
     *            or a valid URL
     * @return a {@link Properties} object fill with properties value found in
     *         the targeted file
     * @throws PEtALSCFException
     *             if error occurs when reading external properties file
     */
    public final static Properties loadProperties(String propertiesFilePath)
            throws PEtALSCFException {
        Properties properties = null;
        if (!StringHelper.isNullOrEmpty(propertiesFilePath)) {
            boolean urlLocation = true;
            URL url = null;
            try {
                url = new URL(propertiesFilePath);
            } catch (MalformedURLException e1) {
                urlLocation = false;
            }
            InputStream propFileStream = null;
            if (urlLocation) {
                propFileStream = loadPropertiesFileFromUrl(url);
            } else {
                propFileStream = loadPropertiesFileFromRelativePath(propertiesFilePath);
            }
            if (propFileStream != null) {
                properties = loadProperties(propFileStream);
            } else {
                throw new PEtALSCFException("Cannot find properties at location : "
                        + propertiesFilePath);
            }
        }
        return properties;
    }

    /**
     * Load {@link Properties} object from a given {@link InputStream}
     *
     * @param propFileStream
     *            a properties file {@link InputStream}
     * @return the loaded {@link Properties} object, null if error occurs during
     *         {@link Properties} loading
     */
    private static final Properties loadProperties(InputStream propFileStream) {
        Properties fileProperties = new Properties();
        try {
            fileProperties.load(propFileStream);
        } catch (IOException e) {
            fileProperties = null;
        }
        return fileProperties;
    }

    private static final InputStream loadPropertiesFileFromRelativePath(String propertiesFilePath) {
        InputStream result = null;
        // try to get the relative path
        String cleanPropertiesFilePath = "";
        if (!propertiesFilePath.startsWith(File.separator)) {
            cleanPropertiesFilePath = File.separator + propertiesFilePath;
        }
        File propFile = new File(System.getenv("PETALS_HOME") + cleanPropertiesFilePath);
        if (propFile.exists()) {
            try {
                return new FileInputStream(propFile);
            } catch (FileNotFoundException e) {
                // Can't be thrown
            }
        }
        return result;
    }

    private static final InputStream loadPropertiesFileFromUrl(URL url) throws PEtALSCFException {
        URLConnection connection = null;
        InputStream is = null;

        try {
            connection = url.openConnection();
            is = connection.getInputStream();
        } catch (IOException e) {
            throw new PEtALSCFException("Can not get InputStream from URL " + url, e);
        }
        return is;
    }

    /**
     * <p>
     * Replace place holders in the toProperties {@link Properties} object by
     * values found in fromProperties {@link Properties} object. Place holders
     * look like : ${property name in fromProperties object}
     * </p>
     * <p>
     * Exemple : <br/> toProperties property place holder : ${foo.bar}<br/>
     * fromProperties property : foo.bar <br/> Property with value ${foo.bar} in
     * toProperties object will be overiden by the value of foo.bar property of
     * the fromProperties object
     * </p>
     *
     * @param fromProperties
     *            properties that will replace place holders in toProperties
     *            object
     * @param toProperties
     *            a {@link Properties} object with some place holder as value
     * @throws PEtALSCFException
     *             if a property specified in toProperties cannot be found in
     *             fromProperties
     */
    public static void fillPlaceHolder(Properties fromProperties, Properties toProperties)
            throws PEtALSCFException {
        if (fromProperties != null && toProperties != null) {
            Set<Entry<Object, Object>> externalizedProperties = new HashSet<Entry<Object, Object>>(
                    toProperties.entrySet());
            for (Entry<Object, Object> entry : externalizedProperties) {
                String entryValue = (String) entry.getValue();
                if (entryValue.startsWith("${") && entryValue.indexOf("}") > 0) {
                    String externalRefKey = entryValue.substring(entryValue.indexOf("${") + 2,
                            entryValue.indexOf("}"));
                    String realValue = fromProperties.getProperty(externalRefKey);
                    if (!StringHelper.isNullOrEmpty(realValue)) {
                        toProperties.setProperty((String) entry.getKey(), realValue);
                    } else {
                        throw new PEtALSCFException("Can't find the following properties '"
                                + externalRefKey + "' in external properties");
                    }
                }
            }
        }
    }

    /**
     * <p>
     * Replace place holders in the toProperties {@link Properties} object by
     * values found in external properties file denoted by the given
     * propertiesFilePath. Place holders look like :
     * ${property.name.in.fromProperties.object}
     * </p>
     * <p>
     * Exemple : <br/> toProperties property place holder : ${foo.bar}<br/>
     * External properties file property : foo.bar <br/> Property with value
     * ${foo.bar} in toProperties object will be overiden by the value of
     * foo.bar property of the external properties file
     * </p>
     *
     * @param propertiesFilePath
     *            The file path can be relative path based on Petals root path
     *            or a valid URL
     *            <p>
     *            Exemples : <br/> Relative path : /conf/external.properties
     *            <br/> Absolute path : file:///C:/conf/external.properties
     *            <br/> URL : ftp://user:pwd@host:port/external.properties
     *            </p>
     * @param toProperties
     *            a {@link Properties} object with some place holder as value
     * @throws PEtALSCFException
     *             if error occurs when reading external properties file or if a
     *             property specified in toProperties cannot be found in
     *             external properties file
     */
    public static void fillPlaceHolder(String propertiesFilePath, Properties toProperties)
            throws PEtALSCFException {
        Properties externalProperties = loadProperties(propertiesFilePath);
        fillPlaceHolder(externalProperties, toProperties);
    }
}
